/* @i
 * File: createIntermediate.c
 *
 * Modified version - all required *.c files included in this
 *
 * Form: C
 * Path: D511.5_Graphs/BalDiSo (Graphs)
 * Note: calculate different tree measures
 * Meta: Sokal, Tree Balance, Tree Imbalance, Phylogeny, Taxa
 * Date: 2012-08-15, Wed
 *       2012-11-29, Thur, copied/adapted from D511.5_Graphs/Sokal/sokal.c
 */

/* @ include files */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

// ensure correct definition of __func__
#if __STDC_VERSION__ < 199901L
# if __GNUC__ >= 2
#  define __func__ __func__
# else
#  define __func__ "<unknown>"
# endif
#endif

// #include "globals.h"
enum {FAILURE = -1, OK};
enum {MAXWORD = 100, MAXLINE=1000, MAXLEN = 10000};
enum {BASE10 = 10};
enum {SPLT = 0, LEAF};

#define MAX(A,B) ((A)>(B)?(A):(B))
#define ABS(A) ((A)>0.00?(A):(-(A)))

typedef struct node {
  int type; /* type of node = {split or leaf} */
  int cnt; /* successive counter of nodes */
  int intr; /* number of internal nodes */
  int subt; /* number of subtended leaves */
  int subtL; /* number of subtended leaves by left branch */
  int subtR; /* number of subtended leaves by right branch*/
  int max; /* maximum of interior nodes subtended by split */
  int leafID; /* identifier for leaf */
  char name[MAXWORD]; /* name of split/leaf */
  struct node *left; /* pointer to left branch */
  struct node *rght; /* pointer to right branch */
  int dia; /* diameter of (sub-)tree */
  int height; /* height of (sub-)tree */
} NODE;


// #include "utils.h"
#define RANDOM_BRANCH 0.12345

int PrintTree(NODE *rSubTree){

  if (NULL == rSubTree->left){
    printf("%s", rSubTree->name);
  }
  else {
    //printf("%d: %p: %p-%p\n", __LINE__, rSubTree, rSubTree->left, rSubTree->rght);
    printf("(");
    PrintTree(rSubTree->left);
    printf(",");
    PrintTree(rSubTree->rght);
    printf("):%f\n", RANDOM_BRANCH);
  }

  return OK;

} /* this is the end of PrintTree() */



// #include "readTree.h"
int ReadTreeFile(char **pcTree_p, char zcInFile[],
		 char zcAux[], char zcError[]){

  /* @<variables local to ReadTreeFile> */
  FILE *pfIn = NULL; 
  char zcScanline[MAXLINE+1] = {0};
  char *pcToken = NULL;
  int iStrlen = 0;
  int iTotLen = 0;

  /* @<open file for reading> */
  if (NULL == (pfIn = fopen(zcInFile, "r"))){
    sprintf(zcError, "%s:%s:%d: could not open file %s for reading\n", 
	    __func__, __FILE__, __LINE__, zcInFile);
    return FAILURE;
  }

  /* @<initialise memory> */
  if (NULL == (*pcTree_p = (char *)malloc(1))){
    sprintf(zcError, "%s:%s:%d: could not malloc 1 char for tree\n",
			__func__, __FILE__, __LINE__);
    return FAILURE;
  }
  else{
    (*pcTree_p)[0] = '\0';
    iTotLen = 1;
  }

  while(NULL != fgets(zcScanline, MAXLINE, pfIn)){

    pcToken = strtok(zcScanline, "\n");
    
    iStrlen = strlen(zcScanline);

    if (NULL == (*pcTree_p = (char *)realloc(*pcTree_p, iTotLen+iStrlen))){
      sprintf(zcError, "%s:%s:%d: could not realloc tree from %d to %d char\n",
	      __func__, __FILE__, __LINE__, iTotLen, iTotLen+iStrlen);
      return FAILURE;
    }
    else{
      strcat(*pcTree_p, zcScanline);
      iTotLen += iStrlen;
    }

  } /* EOF */

  /* @= */
  fclose(pfIn); pfIn = NULL;
  return OK;

} /* this is the end of ReadTreeFile() */



/* @* CreateLL
 */
int CreateLL(NODE **prTree_p, int *iNoof_p, 
	     char *pcTree, char zcAux[], char zcError[]){

  /* @<variables local to CreateLL> */
  int iOpen = 0; /* number of open brackets */
  int iClse = 0; /* number of close brackets */
  int iCmma = 0; /* number of commas */
  int iColn = 0; /* number of colons */
  int iCnt = 0; /* counter of nodes already dished out */
  char *pcPtr = pcTree;
  
  /* @<count tokens to establish number of taxa/nodes>*/
  while ('\0' != *pcPtr){

    switch(*pcPtr){

    case '(':
      iOpen++;
      break;
    case ')':
      iClse++;
      break;
    case ',':
      iCmma++;
      break;
    case ':':
      iColn++;
      break;
    }
    pcPtr++;

  } /* EO string */

  if ( (iOpen != iClse) || (iOpen != iCmma) ){
    sprintf(zcError, 
	    "%s:%s:%d: FATAL -- open=%d, close=%d, comma=%d, colon=%d\n", 
	    __func__, __FILE__, __LINE__, 
	    iOpen, iClse, iCmma, iColn);
    return FAILURE;
  }
  else if (iColn != 2*iOpen){
#ifdef VERBOSE
    printf("%s:%s:%d: WARNING -- open=%d, close=%d, comma=%d, colon=%d\n",
	   __func__, __FILE__, __LINE__,
	   iOpen, iClse, iCmma, iColn);
#endif
  }

  *iNoof_p = iOpen+1;

  /* @<malloc memory for node/taxa list> */
  if (NULL == (*prTree_p = (NODE *)malloc(sizeof(NODE)))){
    sprintf(zcError, "%s:%s:%d: could nor malloc root NODE for tree\n",
	    __func__, __FILE__, __LINE__);
    return FAILURE;
  }
  else{
    memset(*prTree_p, 0, sizeof(NODE));
  }


  (*prTree_p)->type = LEAF;
  (*prTree_p)->cnt = iCnt;
  strcpy((*prTree_p)->name, "root");
  (*prTree_p)->intr = 0;

  /* @<disect string into left/right> */
  if (OK != DisectLR(*prTree_p, &iCnt, pcTree, 
		     zcAux, zcError)){
    sprintf(zcAux, "%s:%s:%d: L/R split failed at root level\n",
	    __func__, __FILE__, __LINE__);
    if (strlen(zcError)+strlen(zcAux)<MAXLEN){
      strcat(zcError, zcAux);
    }
    else {
      printf("%s", zcAux);
    }
    return FAILURE;
  }


  /* @= */
  return OK;

} /* this is the end of CreateLL() */


/* @* DisectLR
 */
int DisectLR(NODE *prTree, int *iCnt_p, char *pcTree, 
	     char zcAux[], char zcError[]){

  /* @<variables local to DisectLR> */
  char *pcLeft = NULL;
  char *pcRght = NULL;
  char *pcIter = NULL;
  int iBrace = 0;
  int iLocal = *iCnt_p;
  static int siLeafCounter = 0;

  /* @<check if leaf> */
  if (NULL == strchr(pcTree, '(')){

    prTree->type = LEAF;
    prTree->cnt = *iCnt_p;
    prTree->leafID = siLeafCounter++;
    prTree->subt = 1;
    strcpy(prTree->name, pcTree);
    prTree->left = NULL;
    prTree->rght = NULL;
    // KB prTree->max = prTree[iLocal].intr;
    prTree->height = 1;
    prTree->dia    = 1;

#ifdef VVVERBOSE
    printf("\n%s is a LEAF", pcTree);
#endif
    return OK;
  }

  /* @<strip outer parenthesis> */
  pcLeft = strchr(pcTree, '(');
  pcLeft++;
  pcIter = strrchr(pcTree, ')');
  *pcIter = '\0';
  /* found left beginning */
#ifdef VVVERBOSE
  printf("\n%s:%s:%d: Left: %s", __func__, __FILE__, __LINE__, pcLeft);
#endif

  /* @<identify end of left term> */
  pcIter = pcLeft;
  iBrace = 0;

  while ('\0' != *pcIter){

    if ('(' == *pcIter){
      iBrace++;
      pcIter++;
    }
    else if (')' == *pcIter){
      iBrace--;
      pcIter++;
    }
    else if ( (0 == iBrace) && (',' == *pcIter)){
      *pcIter = '\0';
      break;
    }
    else {
      pcIter++;
    }
  } /* found left end */
#ifdef VVVERBOSE
  printf("\n%s:%s:%d: Left: %s", __func__, __FILE__, __LINE__, pcLeft);
#endif
  /* @<beginning of right term is just next char> */
  pcRght = pcIter+1;
  /* found right end */
#ifdef VVVERBOSE
  printf("\n%s:%s:%d: Rght: %s\n", __func__, __FILE__, __LINE__, pcRght);
#endif


  /* @<recurse> */
  prTree->type = SPLT;
  prTree->cnt = *iCnt_p;
  sprintf(prTree->name, "split-%d", *iCnt_p);
  (*iCnt_p)++;
  prTree->left = (NODE *)malloc(sizeof(NODE));
  prTree->left->intr = prTree->intr + 1;
  if (OK != DisectLR(prTree->left, &*iCnt_p, pcLeft, zcAux, zcError)){
    sprintf(zcAux, "%s:%s:%d: error occurred during LEFT parsing", 
	    __func__, __FILE__, __LINE__);
    if (strlen(zcError)+strlen(zcAux)<MAXLEN){
      strcat(zcError, zcAux);
    }
    else{
      printf("%s", zcAux);
    }
    return FAILURE;
  }
  (*iCnt_p)++;
  prTree->rght = (NODE *)malloc(sizeof(NODE));
  prTree->rght->intr = prTree->intr + 1;
  if (OK != DisectLR(prTree->rght, &*iCnt_p, pcRght, zcAux, zcError)){
    sprintf(zcAux, "%s:%s:%d: error occurred during RIGHT parsing",
            __func__, __FILE__, __LINE__);
    if (strlen(zcError)+strlen(zcAux)<MAXLEN){
      strcat(zcError, zcAux);
    }
    else{
      printf("%s", zcAux);
    }
    return FAILURE;
  }


  /* @= */
  return OK;

} /* this is the end of DisectLR() */

// #include "prune.h"
int RemoveNode(char zcSeqLabel[], int *iSeqID_p, NODE *prTree, char cLeftRght, 
               char zcAux[], char zcError[]){

  NODE *prIter = NULL;
  NODE *prLast = NULL;
  char zcOtherLeaf[MAXWORD] = {0};

  if ( (NULL == prTree) || (NULL == prTree->left) || (NULL == prTree->rght) ){
    sprintf(zcError, 
            "%s:%s:%d: tree cannot be pruned (t=%p, t->l=%p, t->r=%p)\n",
            __func__, __FILE__, __LINE__, 
            prTree, 
            prTree==NULL?NULL:prTree->left, prTree==NULL?NULL:prTree->rght);
    return FAILURE;
  }

  if ('L' == cLeftRght){

    /* @<go down to the left-most leaf> */
    for (prLast = prTree, prIter = prTree; NULL != prIter->left; 
         prLast = prIter, prIter = prLast->left);
    /* @<get label of left-most leaf> */
    strcpy(zcSeqLabel, prIter->name);
    *iSeqID_p = prIter->leafID;
    /* @<delete the left-most node> */
    free(prIter);
    /* @<go to neighbour of former left-most node> */
    prIter = prLast->rght;
    /* @<transfer info from neighbour of left-most to parent> */
    prLast->left = prIter->left;
    prLast->rght = prIter->rght;
    strcpy(prLast->name, prIter->name);
    prLast->leafID = prIter->leafID;
    /* @<delete neighbour (pointer) to former left-most node> 
       info has been transferred to former parent */
    free(prIter);
  }
  else if ('R' == cLeftRght){

    /* @<go down to the right-most leaf> */
    for (prLast = prTree, prIter = prTree; NULL != prIter->rght; 
         prLast = prIter, prIter = prLast->rght);
    /* @<get label of rght-most leaf> */
    strcpy(zcSeqLabel, prIter->name);
    *iSeqID_p =prIter->leafID;
    /* @<delete the rght-most node> */
    free(prIter);
    /* @<go to neighbour of former rght-most node> */
    prIter = prLast->left;
    /* @<transfer info from neighbour of rght-most to parent> */
    prLast->left = prIter->left;
    prLast->rght = prIter->rght;
    strcpy(prLast->name, prIter->name);
    prLast->leafID = prIter->leafID;
    /* @<delete neighbour (pointer) to former rght-most node> 
       info has been transferred to former parent */
    free(prIter);
  } 
  else {
    sprintf(zcError, "%s:%s:%d: unknown reduction scheme (%c)\n", 
            __func__, __FILE__, __LINE__, cLeftRght);
    return FAILURE;
  }

  /* @= */
  return OK;

} /* this is the end of RemoveNode() */



/* @* RemoveMultipleNodes */
int RemoveMultipleNodes(char ***ppcSeqLabels_p, int **piSeqIDs_p, NODE *prTree, 
                        char zcLeftRght[], int iNoof, 
                        char zcAux[], char zcError[]){

  /* @<variable declaration for RemoveMultipleNodes> */
  int i;
  int iSeqID = -1;

  /* @<malloc memory for sequence label names and IDs> */
  if (NULL == (*piSeqIDs_p = (int *)malloc(iNoof*sizeof(int)))){
    sprintf(zcError, "%s:%s:%d: could not malloc %d int for label-IDs\n",
            __func__, __FILE__, __LINE__, iNoof);
    return FAILURE;
  }
  memset(*piSeqIDs_p, 0, iNoof*sizeof(int));
  if (NULL == (*ppcSeqLabels_p = (char **)malloc(iNoof*sizeof(char *)))){
    sprintf(zcError, "%s:%s:%d: could not malloc %d char* for labels\n",
            __func__, __FILE__, __LINE__, iNoof);
    return FAILURE;
  }
  if (NULL == ((*ppcSeqLabels_p)[0] = 
               (char *)malloc(iNoof*MAXWORD*sizeof(char)))){
    sprintf(zcError, "%s:%s:%d: could not malloc %d char for labels\n",
            __func__, __FILE__, __LINE__, iNoof*MAXWORD);
    return FAILURE;
  }
  else {
    (*ppcSeqLabels_p)[0][0] = '\0';
    for (i = 1; i < iNoof; i++){
      (*ppcSeqLabels_p)[i] = (*ppcSeqLabels_p)[i-1] + MAXWORD;
      (*ppcSeqLabels_p)[i][0] = '\0';
    }
  }

  if (0 == strcmp(zcLeftRght, "L")){
    for (i = 0; i < iNoof; i++){
      if (OK != RemoveNode((*ppcSeqLabels_p)[i], &(*piSeqIDs_p)[i], prTree, 'L', 
                           zcAux, zcError)){
        sprintf(zcAux, "%s:%s:%d: could not remove %dth leave from left\n",
                __func__, __FILE__, __LINE__, i);
        strcat(zcError, zcAux);
        return FAILURE;
      }
    } /* went through multiple removals */
  } /* only removed leaves from left */
  else if (0 == strcmp(zcLeftRght, "R")){
    for (i = 0; i < iNoof; i++){
      if (OK != RemoveNode((*ppcSeqLabels_p)[i], &(*piSeqIDs_p)[i], prTree, 'R', 
                           zcAux, zcError)){
        sprintf(zcAux, "%s:%s:%d: could not remove %dth leave from right\n",
                __func__, __FILE__, __LINE__, i);
        strcat(zcError, zcAux);
        return FAILURE;
      }
    } /* went through multiple removals */
  } /* only removed leaves from right */
  else if (0 == strcmp(zcLeftRght, "LR")){
    for (i = 0; i < iNoof; i++){
      if (OK != RemoveNode((*ppcSeqLabels_p)[i], &(*piSeqIDs_p)[i], prTree, 'L', 
                           zcAux, zcError)){
        sprintf(zcAux, "%s:%s:%d: could not remove %dth leave from left/right\n",
                __func__, __FILE__, __LINE__, i);
        strcat(zcError, zcAux);
        return FAILURE;
      }
      i++;
      if (i >= iNoof){
	break;
      }
      if (OK != RemoveNode((*ppcSeqLabels_p)[i], &(*piSeqIDs_p)[i], prTree, 'R', 
                           zcAux, zcError)){
        sprintf(zcAux, "%s:%s:%d: could not remove %dth leave from right/left\n",
                __func__, __FILE__, __LINE__, i);
        strcat(zcError, zcAux);
        return FAILURE;
      }
    } /* went through multiple removals */
  } /* removed leaves from left/right */
  else {
    sprintf(zcError, "%s:%s:%d: unknown leave removal scheme (%s), only 'L', 'R', 'LR' allowed\n",
	    __func__, __FILE__, __LINE__, zcLeftRght);
    return FAILURE;
  }

  /* @= */
  return OK;
  
} /* this is the end of RemoveMultipleNodes() */


/* @* ReattachLeaves */
int ReattachLeaves(NODE **prTree_p, char **ppcSeqLabels, int *piSeqIDs, int iNoof,
                   char zcAux[], char zcError[]){

  NODE * prOld = NULL;
  int i;


  for (i = 0; i < iNoof; i++){

    prOld = *prTree_p;
    *prTree_p = (NODE *)malloc(sizeof(NODE));
    memset(*prTree_p, 0, sizeof(NODE));
    (*prTree_p)->left = prOld;
    (*prTree_p)->rght = (NODE *)malloc(sizeof(NODE));
    memset((*prTree_p)->rght, 0, sizeof(NODE));
    (*prTree_p)->rght->left = NULL;
    (*prTree_p)->rght->rght = NULL;
    strcpy((*prTree_p)->rght->name, ppcSeqLabels[iNoof-i-1]);
    (*prTree_p)->rght->leafID = piSeqIDs[iNoof-i-1];
  } /* 0 <= i < iNoof */


  /* @= */
  return OK;

} /* this is the end of ReattachLeaves() */


/* @ proto-types */
void ErrorExit(char zcError[]);
int Commandline(int *iKickOff_p, char zcLeftRght[], char zcInTree[], int argc, char *argv[], char zcError[]);


/* @* main */
int main (int argc, char *argv[]){
  
  /* @<variables local to main@> */
  NODE *prTree = NULL;
  int iLeafID = 0; 
  int iNoofLeaves = 0;
  int iKickOff = 0; 
  int *piSeqIDs = NULL;
  int iRead = 0;
  int iNoof = 0;
  char *pcTree = NULL;
  char **ppcSeqLabelsRead = NULL; 
  char **ppcSeqLabelsCut = NULL; 
  char zcInTree[MAXWORD] = {0};
  char zcSeqLabel[MAXWORD] = {0}; 
  char zcLeftRght[MAXWORD] = {0};
  char zcAux[MAXLEN] = {0}; /* error message */
  char zcError[MAXLEN] = {0}; /* error message */
  char zcIdent[] = "@(#)"__FILE__"(D511.5/BDS): construct balanced tree, migrate leaves out";
  double dSachin = -1.0;
  double dColless = -1.0;
  double dInvMax = -1.0;
  double dShannon = -12345.6789;

  /* @<commandline@> */
  if (OK != Commandline(&iKickOff, zcLeftRght, zcInTree, 
			argc, argv, zcError)){
    ErrorExit(zcError);
  }
  
  /* @<read tree> 
     read file into a string */
  if (OK != ReadTreeFile(&pcTree, zcInTree, zcAux, zcError)){
    ErrorExit(zcError);
  }
  /* @<create linked list> */
  if (OK != CreateLL(&prTree, &iNoof, pcTree, zcAux, zcError)){
    ErrorExit(zcError);
  }


  /* @<remove nodes> */
  if (OK != RemoveMultipleNodes(&ppcSeqLabelsCut, &piSeqIDs, prTree, zcLeftRght, iKickOff,
                                zcAux, zcError)){
    ErrorExit(zcError);
  }
#ifdef VERBOSE
  printf("extreme leaves are %s,%s,%s\n", ppcSeqLabelsCut[0],
         ppcSeqLabelsCut[1], ppcSeqLabelsCut[2]);
#endif
#ifdef VVERBOSE
  printf("this is the pruned tree:\n");
  PrintTree(prTree);
#endif

  /* @<re-attach leaves on the outside> */
  if (OK != ReattachLeaves(&prTree, ppcSeqLabelsCut, piSeqIDs, iKickOff, zcAux, zcError)){
    ErrorExit(zcError);
  }

#ifdef VERBOSE
  printf("leaves reattached to pruned tree:\n");
#endif
  PrintTree(prTree);
  printf(";\n");


  return 0;
  
} /* this is the end of main() */


/* @* Commandline */
#define USAGE "Example: %s  3      L       tree.dnd\n" \
  "Usage:  <exec> <kick> <order> <tree>\n" \
  "         |      |      |       |\n" \
  "         |      |      |       +-tree file\n" \
  "         |      |      +-where to remove leaves ('L' left, 'R' right, 'LR' left-right-left-...)\n" \
  "         |      +-number of leaves to be kicked-off\n" \
  "         +-executable\n"

enum {ARG_EXEC = 0, ARG_KICK, ARG_LEFTRGHT, ARG_TREE, ARGC};
int Commandline(int *iKickOff_p, char zcLeftRght[], char zcInTree[], int argc, char *argv[], char zcError[]){

  /* @<variables local to Commandline@> */
  char *pcEnd = NULL;

  /* @<check number of arguments@> */
  if (ARGC != argc){
    sprintf(zcError, "%s:%s:%d: %d args expected, %d supplied\n"USAGE, 
            __func__, __FILE__, __LINE__, ARGC-1, argc-1, argv[0]);
    return FAILURE;
  }

  /* @<number of sequences to be migrated out of (perfect?) tree> */
  *iKickOff_p = (int)strtol(argv[ARG_KICK], &pcEnd, BASE10);
  if ('\0' != *pcEnd){
    sprintf(zcError, "%s:%s:%d: kick-off not clean int (%s)\n"USAGE, 
            __func__, __FILE__, __LINE__, argv[ARG_KICK], argv[0]);
    return FAILURE;
  }

  /* @<copy where to remove leaves> */
  strcpy(zcLeftRght, argv[ARG_LEFTRGHT]);
  if ( (0 != strcmp(zcLeftRght, "L")) && (0 != strcmp(zcLeftRght, "R")) && (0 != strcmp(zcLeftRght, "LR")) ){
    sprintf(zcError, "%s:%s:%d: leaves have to be removed from left (L) or right (R), %s not known\n"USAGE, 
	    __func__, __FILE__, __LINE__, zcLeftRght, argv[0]);
    return FAILURE;
  }

  /* @<copy tree file name> */
  strcpy(zcInTree, argv[ARG_TREE]);


  /* @= */
  return OK;

} /* this is the end of Commandline() */


/* @* ErrorExit */
void ErrorExit(char zcError[]){
  printf("+-----------------------------------+\n"
	  "| Program did NOT run to completion |\n"
	  "+-----------------------------------+\n"
         "%s",
         zcError);
  exit(FAILURE);
} /* this is the end of ErrorExit() */


/*i
 * EOF readRemove.c 
 */
