/* @i
 * File: reshuffleSeq.c
 *
 * Form: C
 * Path: fabian@myosin:~/Omega/Papers/2011-Sep-Scaling/Tools
 * Note: read sequences and re-shuffle order
 * Meta: Fasta, Sequences, Vienna Format
 * Date: 2011-10-17, Mon
 */

/* @ include files */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

/* @ definitions */
enum {UNKNOWN = -1, FAILURE = -1, OK};
enum {MAXWORD = 100, MAXLINE = 100000, MAXLEN = 10000};
enum {BASE10 = 10};

#define  VERBOSE 0
#define VVERBOSE 0

typedef struct {
  char label[MAXWORD];
  char *seq;
} sequence;

typedef struct {
  char label[MAXWORD];
  double w;
} weight;

typedef struct {
  int orig;
  double tag;
  sequence *target;
} randomise;


/* @ prototypes */
int Commandline(int *iSeed_p, char zcVienna[], char zcWeights[], double *dOffset_p, 
		int argc, char *argv[], char zcError[]);
void ErrorExit(char zcError[]);
int ReadViennaSeqs(sequence **prSeqs_p, int *iSeqs_p, char zcVienna[], 
		   char zcAux[], char zcError[]);
int ReadWeights(weight **prWeights_p, 
		int iSeqs, char zcWeights[], char zcAux[], char zcError[]);
int Synchrone(sequence *prSeqs, weight *prWeights, int iSeqs, 
	      char zcAux[], char zcError[]);
int ReshuffleSeqs(randomise **prRndPtr_p, int iSeed, 
		  sequence *prSeqs, weight *prWeights, int iSeqs, double dOffset, 
		  char zcAux[], char zcError[]);
int CompRndAsc(const void *cv1, const void *cv2);
int PrintSequences(randomise *prRndPtr, sequence *prSeqs, int iSeqs,
		   char zcAux[], char zcError[]);



/* @* main */
int main (int argc, char *argv[]){

  /* @<variables local to main@> */
  int iSeed = 0; 
  double dOffset = 0.00;
  char zcVienna[MAXWORD] = {0};
  char zcWeights[MAXWORD] = {0};
  char zcAux[MAXLEN] = {0}; /* error message */
  char zcError[MAXLEN] = {0}; /* error message */
  char zcIdent[] = "@(#)"__FILE__": read and reshuffle order of sequences";
  int iSeqs = 0;
  sequence *prSeqs = NULL;
  weight *prWeights = NULL;
  randomise *prRndPtr = NULL;


  /* @<commandline@> */
  if (OK != Commandline(&iSeed, zcVienna, zcWeights, &dOffset, 
			argc, argv, zcError)){
    ErrorExit(zcError);
  }
  if (VVERBOSE){
    printf("vienna=%s, weights=%s, seed=%d\n", zcVienna, zcWeights, iSeed);
  }

  /* @<read sequences> */
  if (OK != ReadViennaSeqs(&prSeqs, &iSeqs, zcVienna, zcAux, zcError)){
    ErrorExit(zcError);
  }
  if (VERBOSE){
    int i;
    printf("there are %d sequences\n", iSeqs);

    if (VVERBOSE){
      for (i = 0; i < iSeqs; i++){
	printf("@%s#%s\n", prSeqs[i].label, prSeqs[i].seq);
      }
    }
  }

  /* @<read weights> */
  if (OK != ReadWeights(&prWeights, iSeqs, zcWeights, zcAux, zcError)){
    ErrorExit(zcError);
  }
  if (VVERBOSE){
    int i;

    for (i = 0; i < iSeqs; i++){
      printf("%f\t%s\t%s\n", prWeights[i].w, prWeights[i].label, prSeqs[i].label);
    }
  }


  /* @<check synchronicity> */
  if (OK != Synchrone(prSeqs, prWeights, iSeqs, zcAux, zcError)){
    ErrorExit(zcError);
  }

  /* @<re-shuffle sequences> */
  if (OK != ReshuffleSeqs(&prRndPtr, iSeed, prSeqs, prWeights, iSeqs, dOffset, 
			  zcAux, zcError)){
    ErrorExit(zcError);
  }
  if (VVERBOSE && (NULL != prWeights) ){
    int i;

    for (i = 0; i < iSeqs; i++){
      printf("%f\t%f\n", prRndPtr[i].tag, prWeights[prRndPtr[i].orig].w);
    }
  }


  /* @<print re-ordered sequence file> */
  if (OK != PrintSequences(prRndPtr, prSeqs, iSeqs,
			   zcAux, zcError)){
    ErrorExit(zcError);
  }



  /* @i 
   * prepare for exit */
  if (VERBOSE){
    printf("+---------------------------+\n"
	   "| Program ran to completion |\n"
	   "+---------------------------+\n"); 
  }
  /* @= */ 
  return OK;

} /* this is the end of main() */




/* @* PrintSequences */
int PrintSequences(randomise *prRndPtr, sequence *prSeqs, int iSeqs,
		   char zcAux[], char zcError[]){

  int i;
  FILE *pfOut = NULL;
  char zcOutfile[] = "reorderedSeqs.fa";

  /* @<open out-file> */
  if (NULL == (pfOut = fopen(zcOutfile, "w"))){
    sprintf(zcError, "%s:%s:%d: could not open file %s for writing\n",
	    __FUNCTION__, __FILE__, __LINE__, zcOutfile);
    return FAILURE;
  }


  /* @<write sequences> */
  for (i = 0; i < iSeqs; i++){

    fprintf(pfOut, ">%s\n%s\n", 
	    prSeqs[prRndPtr[i].orig].label, 
	    prSeqs[prRndPtr[i].orig].seq);

  } /* 0 <= i < iSeqs */


  /* @= */
  fclose(pfOut); pfOut = NULL;
  return OK;

} /* this is the end of PrintSequences() */

/* @* ReshuffleSeqs */
int ReshuffleSeqs(randomise **prRndPtr_p, int iSeed, 
		  sequence *prSeqs, weight *prWeights, int iSeqs, double dOffset, 
		  char zcAux[], char zcError[]){

  int i;
  double dStep = dOffset / iSeqs;


  /* @<allocate memory for auxilliary structure> */
  if (NULL == (*prRndPtr_p = (randomise *)malloc(iSeqs * sizeof(randomise)))){
    sprintf(zcError, "%s:%s:%d: could not malloc %d randomise\n",
	    __FUNCTION__, __FILE__, __LINE__, iSeqs);
    return FAILURE;
  }

  /* @<seed random number generator> */
  srand48(iSeed);

  /* @<initialise randomised structures> */
  for (i = 0; i < iSeqs; i++){

    (*prRndPtr_p)[i].orig   = i;
    (*prRndPtr_p)[i].target = &(prSeqs[i]);
    if (NULL != prWeights){
      (*prRndPtr_p)[i].tag = prWeights[i].w;
    }
    else if (0.00 != dOffset){
      (*prRndPtr_p)[i].tag = i * dStep;
    }
    else {
      (*prRndPtr_p)[i].tag = 0.00;
    }
  } /* 0 <= i < iSeqs */

  /* @<add random number to tag> */
  for (i = 0; i < iSeqs; i++){

    double dRand = drand48();

    (*prRndPtr_p)[i].tag += dRand;

  } /* @<add random number to tag> */


  /* @<re-order pointers> */
  qsort(*prRndPtr_p, iSeqs, sizeof(randomise), CompRndAsc);


  /* @= */
  return OK;

} /* this is the end of ReshuffleSeqs() */

/* @* CompRndAsc */
int CompRndAsc(const void *cv1, const void *cv2){

  randomise rR1 = *(randomise *)cv1;
  randomise rR2 = *(randomise *)cv2;

  if      (rR1.tag < rR2.tag) { return -1; }
  else if (rR1.tag > rR2.tag) { return +1; }
  else                { return  0; }

} /* this is the end of CompRndAsc() */


/* @* Synchrone */
int Synchrone(sequence *prSeqs, weight *prWeights, int iSeqs, 
	      char zcAux[], char zcError[]){

  int i;

  if (NULL == prWeights){
    return OK;
  }

  for (i = 0; i < iSeqs; i++){
    if (0 != strcmp(prSeqs[i].label, prWeights[i].label)){
      sprintf(zcError, "%s:%s:%d: labels for seq %d do not agree (%s / %s)\n",
	      __FUNCTION__, __FILE__, __LINE__, i, 
	      prSeqs[i].label, prWeights[i].label);
      return FAILURE;
    }
  }

  /* @= */
  return OK;

} /*this is the end of Synchrone() */

/* @* ReadWeights */
int ReadWeights(weight **prWeights_p, 
		int iSeqs, char zcWeights[], char zcAux[], char zcError[]){

  FILE *pfWgt = NULL;
  char zcScanline[MAXLINE] = {0};
  char *pcToken = NULL;
  char *pcEnd = NULL;
  int iLines = 0;


  if ('\0' == zcWeights[0]){
    return OK;
  }


  if (NULL == (pfWgt = fopen(zcWeights, "r"))){
    sprintf(zcError, "%s:%s:%d: could not open file %s for reading\n",
            __FUNCTION__, __FILE__, __LINE__, zcWeights);
    return FAILURE;
  }


  if (NULL == (*prWeights_p = (weight *)malloc(iSeqs * sizeof(weight)))){
    sprintf(zcError, "%s:%s:%d: could not malloc %d weights\n",
            __FUNCTION__, __FILE__, __LINE__, iSeqs);
    return FAILURE;
  }


  while (NULL != fgets(zcScanline, MAXLINE, pfWgt)){

    if (NULL == (pcToken = strtok(zcScanline, " \040\t\n"))){
      sprintf(zcError, "%s:%s:%d: no valid token in line %d\n",
	      __FUNCTION__, __FILE__, __LINE__, iLines);
      return FAILURE;
    }
    strcpy((*prWeights_p)[iLines].label, pcToken);

    if (NULL ==(pcToken = strtok(NULL, " \040\t\n"))){
      sprintf(zcError, "%s:%s:%d: no valid distance token in line %d\n",
              __FUNCTION__, __FILE__, __LINE__, iLines);
      return FAILURE;
    }
    (*prWeights_p)[iLines].w = (double)strtod(pcToken, &pcEnd);
    if ('\0' != *pcEnd){
      sprintf(zcError, "%s:%s:%d: distance token in line %d not clean double (%s)\n",
              __FUNCTION__, __FILE__, __LINE__, iLines, pcToken);
      return FAILURE;
    }


    iLines++;

  } /* !EOF */

  /* @= */
  fclose(pfWgt); pfWgt = NULL;
  return OK;

} /* this is the end of ReadWeights() */

/* @* ReadViennaSeqs */
int ReadViennaSeqs(sequence **prSeqs_p, int *iSeqs_p, char zcVienna[], 
		   char zcAux[], char zcError[]){

  FILE *pfVie = NULL; 
  char zcScanline[MAXLINE+1] = {0};
  char *pcToken = NULL;

  /* @<open vienna file> */
  if (NULL == (pfVie = fopen(zcVienna, "r"))){
    sprintf(zcError, "%s:%s:%d: could not open file %s for reading\n",
	    __FUNCTION__, __FILE__, __LINE__, zcVienna);
    return FAILURE;
  }
  
  /* @<read file line by line> */
  *iSeqs_p = 0;
  while(NULL != fgets(zcScanline, MAXLINE, pfVie)){
    
    
    if (NULL == (*prSeqs_p = (sequence *)
		 realloc(*prSeqs_p, (*iSeqs_p+1)*sizeof(sequence)))){
      sprintf(zcError, "%s:%s:%d: could not realloc sequences to %d\n",
	      __FUNCTION__, __FILE__, __LINE__, *iSeqs_p+1);
      return FAILURE;
    }
    
    /* first line is label */
    pcToken = zcScanline;

    if ('>' != *pcToken){
      sprintf(zcError, "%s:%s:%d: missing '>' in sequence %d\n", 
	      __FUNCTION__, __FILE__, __LINE__, *iSeqs_p+1);
      return FAILURE;
    }
    pcToken++;
    /*if (NULL == (pcToken = strtok(pcToken, " \040\t\n"))){
      sprintf(zcError, "%s:%s:%d: no sequence label for sequence %d",
	      __FUNCTION__, __FILE__, __LINE__, *iSeqs_p+1);
      return FAILURE;
    }*/
    if (NULL == (pcToken = strtok(pcToken, "\n"))){
      sprintf(zcError, "%s:%s:%d: no sequence label for sequence %d",
	      __FUNCTION__, __FILE__, __LINE__, *iSeqs_p+1);
      return FAILURE;
    }
    strcpy((*prSeqs_p)[*iSeqs_p].label, pcToken);

    /* read 2nd line (sequence) */
    if (NULL == fgets(zcScanline, MAXLINE, pfVie)){
      sprintf(zcError, "%s:%s:%d: no sequence info for sequence %d\n",
              __FUNCTION__, __FILE__, __LINE__, *iSeqs_p+1);
      return FAILURE;
    }
    
    if (NULL == ((*prSeqs_p)[*iSeqs_p].seq = (char *)
		 malloc(strlen(zcScanline)+1))){
      sprintf(zcError, "%s:%s:%d: could not malloc %d char for sequence %d",
              __FUNCTION__, __FILE__, __LINE__, (int)strlen(zcScanline)+1, *iSeqs_p+1);
      return FAILURE;
    }
    if (NULL == (pcToken = strtok(zcScanline, " \040\t\n"))){
      sprintf(zcError, "%s:%s:%d: no sequence info for sequence %d",
	      __FUNCTION__, __FILE__, __LINE__, *iSeqs_p+1);
      return FAILURE;
    }

    strcpy((*prSeqs_p)[*iSeqs_p].seq, pcToken);


    (*iSeqs_p)++;
  } /* !EOF */

  /* @= */
  fclose(pfVie); pfVie = NULL;
  return OK;

} /* this is the end of ReadViennaSeqs() */




/* @* Commandline */
int Commandline(int *iSeed_p, char zcVienna[], char zcWeights[], double *dOffset_p, 
		int argc, char *argv[], char zcError[]){

  /* @<variables local to Commandline@> */
  int i;
  char *pcEnd = NULL;
  char zcUsage[] = 
    "Example: ./xR -v=globin.fa -s=5   -o=1.234\n"
    "Usage:  <exec> <vienna>    <seed> <offset>\n"
    "         |     |           |      |\n"
    "         |     |           |      +-offset of random weights (optional, default=0.00)\n"
    "         |     |           +-random seed optional, default 0)\n"
    "         |     +-file in vienna format to be reshuffled (mandatory)\n"
    "         +-executable\n";

  *iSeed_p = UNKNOWN;
  zcVienna[0] = '\0';
  zcWeights[0] = '\0';

  for (i = 1; i < argc; i++){

    if (argv[i] == strstr(argv[i], "-v=")){
      strcpy(zcVienna, argv[i]+3);
    }
    else if (argv[i] == strstr(argv[i], "-w=")){
      strcpy(zcWeights, argv[i]+3);
    }
    else if (argv[i] == strstr(argv[i], "-s=")){
      /* @<convert integer@> */
      *iSeed_p = (int)strtol(argv[i]+3, &pcEnd, BASE10);
      if ('\0' != *pcEnd){
	sprintf(zcError, "%s:%d: seed not clean int (%s)\n%s", 
		__FILE__, __LINE__, argv[i]+3, zcUsage);
	return FAILURE;
      }
    }
    else if (argv[i] == strstr(argv[i], "-o=")){
      /* @<convert double@> */
      *dOffset_p = (double)strtod(argv[i]+3, &pcEnd);
      if ('\0' != *pcEnd){
	sprintf(zcError, "%s:%d: offset not clean double (%s)\n%s", 
		__FILE__, __LINE__, argv[i]+3, zcUsage);
	return FAILURE;
      }
    }
    else {
      printf("unknown commandline flag (%s)\n%s", argv[i], zcUsage);
    }

  } /* 1 <= i < argc */


  if ('\0' == zcVienna[0]){
    sprintf(zcError, "%s:%s:%d: no vienna file, mandatory!\n%s", 
	    __FUNCTION__, __FILE__, __LINE__, zcUsage);
    return FAILURE;
  }

  /* @= */
  return OK;

} /* this is the end of Commandline() */


/* @* ErrorExit */
void ErrorExit(char zcError[]){
  printf("+-----------------------------------+\n"
	 "| Program did NOT run to completion |\n"
	 "+-----------------------------------+\n"
         "%s",
         zcError);
  exit(FAILURE);
} /* this is the end of ErrorExit() */




/* @i
 * EOF 
 */
