#!/bin/bash
#
# Test script to run clustal, mafft and muscle 
# with default, balanced and imbalanced trees of varying sizes.
# The quality of the alignment is checked with qscore,
# and these values are collated for each alignment program,
# guide tree and dataset size.
# 

# housekeeping
# define the ranges for the scripte
TREE_DEPTH_FROM=4
TREE_DEPTH_TO=10
INIT_TREE_SIZE=8  # must be set to 2^(TREE_DEPTH_FROM-1)
SAMPLE_FROM=1
SAMPLE_TO=10

# define locations of required programs
# change for your environment
CLUSTAL=...                     # location of Clustal Omega binary
MAFFT=...                       # location of Mafft binary
MUSCLE=...                      # location of Muscle binary
RESHUFFLE=reshuffleSeq          # location of reshuffleSeq binary
CONSTRUCTTREE=constructTree     # location of constructTree binary
NEWICKTOMAFFT=newick2mafft.rb   # location of newick2mafft.rb ruby script 
                                #   (download from http://mafft.cbrc.jp/alignment/software/treein.html)

# reference files
BASE_SEQS=sdr.vie
BASE_REFS_IN=sdr-ref-in.vie
BASE_REFS_OUT=sdr-ref-out.vie
NO_BASE_REFS=13

# create a subdirectory to store the results
RESULTS_DIR=results
mkdir $RESULTS_DIR

TREE_SIZE=$INIT_TREE_SIZE
for TREE_DEPTH in $(seq $TREE_DEPTH_FROM $TREE_DEPTH_TO)
do 
	# update the tree size
	((TREE_SIZE=TREE_SIZE*2))	
	
	# create sub-directory for this tree size and copy files into it
	SUB_DIR=$TREE_SIZE
	mkdir $SUB_DIR
	cp $BASE_SEQS $SUB_DIR/.
	cp $BASE_REFS_IN $SUB_DIR/.
	cp $BASE_REFS_OUT $SUB_DIR/.
	cd $SUB_DIR
	
	# define output files and their extensions
	CLOUT=cl-qscore-$TREE_SIZE
	MAOUT=ma-qscore-$TREE_SIZE
	MUOUT=mu-qscore-$TREE_SIZE
	EXT_DEF=def
	EXT_BAL=rbal
	EXT_CHAIN=rimb

	# create mafft balanced and chained trees, 
	# as these will not change across the samples for this tree size
	if [ -f tmp.txt ]
	then
		rm tmp.txt
	fi
	for i in $(seq 1 $TREE_SIZE)
   	do 
		echo $i >> tmp.txt
   	done
	$CONSTRUCTTREE $TREE_DEPTH tmp.txt
	SCALE=$(echo "scale=6; 1/($TREE_SIZE*10)" | bc)
	ruby $NEWICKTOMAFFT $SCALE tmp-bal.dnd > tree-bal-$TREE_SIZE.mafft 2> /dev/null
	ruby $NEWICKTOMAFFT $SCALE tmp-imb.dnd > tree-imb-$TREE_SIZE.mafft 2> /dev/null

	# process samples for this tree size
	for N in $(seq $SAMPLE_FROM $SAMPLE_TO)
	do
		# create temporary randomised input sequence file (incl ref seqs in)
		# first, reshuffle the sequences, using N as the prng seed
		$RESHUFFLE -v=$BASE_SEQS -s=$N
		# determine how many 'normal' records to include
		((NORMAL_RECS=($TREE_SIZE-$NO_BASE_REFS)*2))
		# extract these records from the shuffled sequences
		head -$NORMAL_RECS reorderedSeqs.fa > tmp.vie
		# catenate the input references and the extracted records
		cat $BASE_REFS_IN  tmp.vie > in.vie
		# reshuffle input sequences again
		$RESHUFFLE -v=in.vie -s=$N
		# rename the output file
		mv reorderedSeqs.fa in.vie

		# create the balanced and imbalanced trees
		# extract labels
		grep '>' in.vie | sed 's/>//g' > tmp.txt
		# create trees
		$CONSTRUCTTREE $TREE_DEPTH tmp.txt


		#
		# Clustal runs
		#

		# default tree
		$CLUSTAL --threads=1 -i in.vie -o out.vie --force 
		# qscore score for aligned sequences
		echo $N " " $(qscore -test out.vie -ref $BASE_REFS_OUT) >> $CLOUT.$EXT_DEF
		rm out.vie

		# balanced tree
		$CLUSTAL --threads=1 -i in.vie -o out.vie --force --guidetree-in=tmp-bal.dnd 
		# qscore score for aligned sequences
		echo $N " " $(qscore -test out.vie -ref $BASE_REFS_OUT) >> $CLOUT.$EXT_BAL
		rm out.vie

		# chained tree 
		$CLUSTAL --threads=1 -i in.vie -o out.vie --force --guidetree-in=tmp-imb.dnd 
		# qscore score for aligned sequences
		echo $N " " $(qscore -test out.vie -ref $BASE_REFS_OUT) >> $CLOUT.$EXT_CHAIN
		rm out.vie


		#
		# Mafft runs
		#

		# default tree
		$MAFFT --thread 1 --anysymbol --unweight in.vie > out.vie 2> /dev/null
		# qscore score for aligned sequences
		echo $N " " $(qscore -test out.vie -ref $BASE_REFS_OUT) >> $MAOUT.$EXT_DEF
		rm out.vie

		# balanced tree
		$MAFFT --thread 1 --treein tree-bal-$TREE_SIZE.mafft --anysymbol --unweight in.vie > out.vie 2> /dev/null
		# qscore score for aligned sequences
		echo $N " " $(qscore -test out.vie -ref $BASE_REFS_OUT) >> $MAOUT.$EXT_BAL
		rm out.vie

		# chained tree 
		$MAFFT --thread 1 --treein tree-imb-$TREE_SIZE.mafft --anysymbol --unweight in.vie > out.vie 2> /dev/null
		# qscore score for aligned sequences
		echo $N " " $(qscore -test out.vie -ref $BASE_REFS_OUT) >> $MAOUT.$EXT_CHAIN
		rm out.vie


		#
		# Muscle 2i run
		#

		# default tree
		$MUSCLE -in in.vie -out out.vie -maxiters 2 -quiet
		# qscore score for aligned sequences
		echo $N " " $(qscore -test out.vie -ref $BASE_REFS_OUT) >> $MUOUT.$EXT_DEF
		rm out.vie

		# balanced tree
		$MUSCLE -in in.vie -out out.vie -usetree_nowarn tmp-bal.dnd -maxiters 2 -quiet
		# qscore score for aligned sequences
		echo $N " " $(qscore -test out.vie -ref $BASE_REFS_OUT) >> $MUOUT.$EXT_BAL
		rm out.vie

		# chained tree
		$MUSCLE -in in.vie -out out.vie -usetree_nowarn tmp-imb.dnd -maxiters 2 -quiet
		# qscore score for aligned sequences
		echo $N " " $(qscore -test out.vie -ref $BASE_REFS_OUT) >> $MUOUT.$EXT_CHAIN
		rm out.vie

	done # end of FOR N loop

    # move output files into results directory
	mv $CLOUT.* ../$RESULTS_DIR/.
	mv $MAOUT.* ../$RESULTS_DIR/.
	mv $MUOUT.* ../$RESULTS_DIR/.

	# clean up
	rm *
	cd ..
	rmdir $SUB_DIR

done # end of TREE_DEPTH loop
